// Copyright (C) 1999-2000 Id Software, Inc.

//

#include "g_local.h"



#define MISSILE_PRESTEP_TIME    50

/* Alan! Prototype */
void alanModifyPath( gentity_t *ent );
void weaponSine( gentity_t *ent, float base_speed, float param );

/* Extern console variables.  See g_main.c for a description */
extern vmCvar_t g_rocket_speed;
extern vmCvar_t g_bfg_speed;
extern vmCvar_t g_plasma_speed;
extern vmCvar_t g_grenade_speed;

extern vmCvar_t g_rocket_lifespan;
extern vmCvar_t g_bfg_lifespan;
extern vmCvar_t g_plasma_lifespan;
extern vmCvar_t g_grenade_lifespan;

extern vmCvar_t g_rocket_damage;
extern vmCvar_t g_bfg_damage;
extern vmCvar_t g_plasma_damage;
extern vmCvar_t g_grenade_damage;

extern vmCvar_t g_rocket_splashdamage;
extern vmCvar_t g_bfg_splashdamage;
extern vmCvar_t g_plasma_splashdamage;
extern vmCvar_t g_grenade_splashdamage;

extern vmCvar_t g_rocket_splashradius;
extern vmCvar_t g_bfg_splashradius;
extern vmCvar_t g_plasma_splashradius;
extern vmCvar_t g_grenade_splashradius;


extern vmCvar_t g_rocket_pathparam1;
extern vmCvar_t g_bfg_pathparam1;
extern vmCvar_t g_plasma_pathparam1;
extern vmCvar_t g_grenade_pathparam1;

extern vmCvar_t g_default_weapon_pathparam1;

extern vmCvar_t g_rocket_pathfunc;
extern vmCvar_t g_bfg_pathfunc;
extern vmCvar_t g_plasma_pathfunc;
extern vmCvar_t g_grenade_pathfunc;

extern vmCvar_t g_default_weapon_pathfunc;


/* End Alan! */

/*
================
G_BounceMissile

  ================
*/
void G_BounceMissile( gentity_t *ent, trace_t *trace ) {
	vec3_t  velocity;
	float   dot;
	int             hitTime;
	
	// reflect the velocity on the trace plane
	hitTime = level.previousTime + ( level.time - level.previousTime ) * trace->fraction;
	BG_EvaluateTrajectoryDelta( &ent->s.pos, hitTime, velocity );
	dot = DotProduct( velocity, trace->plane.normal );
	VectorMA( velocity, -2*dot, trace->plane.normal, ent->s.pos.trDelta );
	
	if ( ent->s.eFlags & EF_BOUNCE_HALF ) {
		VectorScale( ent->s.pos.trDelta, 0.65, ent->s.pos.trDelta );
		// check for stop
		if ( trace->plane.normal[2] > 0.2 && VectorLength( ent->s.pos.trDelta ) < 40 ) {
			G_SetOrigin( ent, trace->endpos );
			return;
		}
	}
	
	VectorAdd( ent->r.currentOrigin, trace->plane.normal, ent->r.currentOrigin);
	VectorCopy( ent->r.currentOrigin, ent->s.pos.trBase );
	ent->s.pos.trTime = level.time;
}


/*
================
G_MissileImpact

================
*/
void G_MissileImpact( gentity_t *ent, trace_t *trace ) {
	gentity_t               *other;
	qboolean                hitClient = qfalse;
	
	other = &g_entities[trace->entityNum];
	
	// check for bounce
	if ( !other->takedamage &&
		( ent->s.eFlags & ( EF_BOUNCE | EF_BOUNCE_HALF ) ) ) {
		G_BounceMissile( ent, trace );
		G_AddEvent( ent, EV_GRENADE_BOUNCE, 0 );
		return;
	}
	
	// impact damage
	if (other->takedamage) {
		// FIXME: wrong damage direction?
		if ( ent->damage ) {
			vec3_t  velocity;
			
			if( LogAccuracyHit( other, &g_entities[ent->r.ownerNum] ) ) {
				g_entities[ent->r.ownerNum].client->ps.persistant[PERS_ACCURACY_HITS]++;
				hitClient = qtrue;
			}
			BG_EvaluateTrajectoryDelta( &ent->s.pos, level.time, velocity );
			if ( VectorLength( velocity ) == 0 ) {
				velocity[2] = 1;        // stepped on a grenade
			}
			G_Damage (other, ent, &g_entities[ent->r.ownerNum], velocity,
				ent->s.origin, ent->damage, 
				0, ent->methodOfDeath);
		}
	}
	
	if (!strcmp(ent->classname, "hook")) {
		gentity_t *nent;
		vec3_t v;
		
		nent = G_Spawn();
		if ( other->takedamage && other->client ) {
			
			G_AddEvent( nent, EV_MISSILE_HIT, DirToByte( trace->plane.normal ) );
			nent->s.otherEntityNum = other->s.number;
			
			ent->enemy = other;
			
			v[0] = other->r.currentOrigin[0] + (other->r.mins[0] + other->r.maxs[0]) * 0.5;
			v[1] = other->r.currentOrigin[1] + (other->r.mins[1] + other->r.maxs[1]) * 0.5;
			v[2] = other->r.currentOrigin[2] + (other->r.mins[2] + other->r.maxs[2]) * 0.5;
			
			SnapVectorTowards( v, ent->s.pos.trBase );      // save net bandwidth
		} else {
			VectorCopy(trace->endpos, v);
			G_AddEvent( nent, EV_MISSILE_MISS, DirToByte( trace->plane.normal ) );
			ent->enemy = NULL;
		}
		
		SnapVectorTowards( v, ent->s.pos.trBase );      // save net bandwidth
		
		nent->freeAfterEvent = qtrue;
		// change over to a normal entity right at the point of impact
		nent->s.eType = ET_GENERAL;
		ent->s.eType = ET_GRAPPLE;
		
		G_SetOrigin( ent, v );
		G_SetOrigin( nent, v );
		
		ent->think = Weapon_HookThink;
		ent->nextthink = level.time + FRAMETIME;
		
		ent->parent->client->ps.pm_flags |= PMF_GRAPPLE_PULL;
		VectorCopy( ent->r.currentOrigin, ent->parent->client->ps.grapplePoint);
		
		trap_LinkEntity( ent );
		trap_LinkEntity( nent );
		
		return;
	}
	
	// is it cheaper in bandwidth to just remove this ent and create a new
	// one, rather than changing the missile into the explosion?
	
	if ( other->takedamage && other->client ) {
		G_AddEvent( ent, EV_MISSILE_HIT, DirToByte( trace->plane.normal ) );
		ent->s.otherEntityNum = other->s.number;
	} else {
		G_AddEvent( ent, EV_MISSILE_MISS, DirToByte( trace->plane.normal ) );
	}
	
	ent->freeAfterEvent = qtrue;
	
	// change over to a normal entity right at the point of impact
	ent->s.eType = ET_GENERAL;
	
	SnapVectorTowards( trace->endpos, ent->s.pos.trBase );  // save net bandwidth
	
	G_SetOrigin( ent, trace->endpos );
	
	// splash damage (doesn't apply to person directly hit)
	if ( ent->splashDamage ) {
		if( G_RadiusDamage( trace->endpos, ent->parent, ent->splashDamage, ent->splashRadius, 
			other, ent->splashMethodOfDeath ) ) {
			if( !hitClient ) {
				g_entities[ent->r.ownerNum].client->ps.persistant[PERS_ACCURACY_HITS]++;
			}
		}
	}
	
	trap_LinkEntity( ent );
}

/*
================
G_ExplodeMissile

Explode a missile without an impact
================
*/
void G_ExplodeMissile( gentity_t *ent ) {
	vec3_t          dir;
	vec3_t          origin;
	
	BG_EvaluateTrajectory( &ent->s.pos, level.time, origin );
	SnapVector( origin );
	G_SetOrigin( ent, origin );
	
	// we don't have a valid direction, so just point straight up
	dir[0] = dir[1] = 0;
	dir[2] = 1;
	
	ent->s.eType = ET_GENERAL;
	G_AddEvent( ent, EV_MISSILE_MISS, DirToByte( dir ) );
	
	ent->freeAfterEvent = qtrue;
	
	// splash damage
	if ( ent->splashDamage ) {
		if( G_RadiusDamage( ent->r.currentOrigin, ent->parent, ent->splashDamage, ent->splashRadius, NULL
			, ent->splashMethodOfDeath ) ) {
			g_entities[ent->r.ownerNum].client->ps.persistant[PERS_ACCURACY_HITS]++;
		}
	}
	
	trap_LinkEntity( ent );
}


/*
================
G_RunMissile

  ================
*/
void G_RunMissile( gentity_t *ent ) {
	vec3_t          origin;
	trace_t         tr;
	
	// get current position
	BG_EvaluateTrajectory( &ent->s.pos, level.time, origin );
	
	// trace a line from the previous position to the current position,
	// ignoring interactions with the missile owner
	trap_Trace( &tr, ent->r.currentOrigin, ent->r.mins, ent->r.maxs, origin, 
		ent->r.ownerNum, ent->clipmask );
	
	VectorCopy( tr.endpos, ent->r.currentOrigin );
	
	if ( tr.startsolid ) {
		tr.fraction = 0;
	}
	
	trap_LinkEntity( ent );
	
	if ( tr.fraction != 1 ) {
		// never explode or bounce on sky
		if ( tr.surfaceFlags & SURF_NOIMPACT ) {
			// If grapple, reset owner
			if (ent->parent && ent->parent->client->hook == ent)
				ent->parent->client->hook = NULL;
			G_FreeEntity( ent );
			return;
		}
		
		G_MissileImpact( ent, &tr );
		if ( ent->s.eType != ET_MISSILE ) {
			return;         // exploded
		}
	}
	
	/* Alan! Missile Modifying Functions */
	
	alanModifyPath( ent );
	
	/* End Alan! */
	
	// check think function after bouncing
	G_RunThink( ent );
}


//=============================================================================

/*
=================
fire_plasma

  =================
*/
gentity_t *fire_plasma (gentity_t *self, vec3_t start, vec3_t dir) {
	gentity_t       *bolt;
	
	VectorNormalize (dir);
	
	bolt = G_Spawn();
	bolt->classname = "plasma";
	bolt->nextthink = level.time + g_plasma_lifespan.value; // Alan! Use _lifespan.value instead of a constant!
	bolt->think = G_ExplodeMissile;
	bolt->s.eType = ET_MISSILE;
	bolt->r.svFlags = SVF_USE_CURRENT_ORIGIN;
	bolt->s.weapon = WP_PLASMAGUN;
	bolt->r.ownerNum = self->s.number;
	bolt->parent = self;
	bolt->damage = g_plasma_damage.value; // Alan! using variables for funness...
	bolt->splashDamage = g_plasma_splashdamage.value; // Alan!
	bolt->splashRadius = g_plasma_splashradius.value; // Alan!
	bolt->methodOfDeath = MOD_PLASMA;
	bolt->splashMethodOfDeath = MOD_PLASMA_SPLASH;
	bolt->clipmask = MASK_SHOT;
	
	bolt->s.pos.trType = TR_LINEAR;
	bolt->s.pos.trTime = level.time - MISSILE_PRESTEP_TIME;         // move a bit on the very first frame
	VectorCopy( start, bolt->s.pos.trBase );
	/* Alan! change from 2k to g_plasma_speed.value */
	VectorScale( dir, g_plasma_speed.value, bolt->s.pos.trDelta );
	/* End Alan! change */
	SnapVector( bolt->s.pos.trDelta );                      // save net bandwidth
	
	VectorCopy (start, bolt->r.currentOrigin);
	
	return bolt;
}       

//=============================================================================


/*
=================
fire_grenade
=================
*/
gentity_t *fire_grenade (gentity_t *self, vec3_t start, vec3_t dir) {
	gentity_t       *bolt;
	
	VectorNormalize (dir);
	
	bolt = G_Spawn();
	bolt->classname = "grenade";
	bolt->nextthink = level.time + g_grenade_lifespan.value; // Alan! Use _lifespan.value instead of a constant!
	bolt->think = G_ExplodeMissile;
	bolt->s.eType = ET_MISSILE;
	bolt->r.svFlags = SVF_USE_CURRENT_ORIGIN;
	bolt->s.weapon = WP_GRENADE_LAUNCHER;
	bolt->s.eFlags = EF_BOUNCE_HALF;
	bolt->r.ownerNum = self->s.number;
	bolt->parent = self;
	bolt->damage = g_grenade_damage.value; // Alan! using variables for funness...
	bolt->splashDamage = g_grenade_splashdamage.value; // Alan!
	bolt->splashRadius = g_grenade_splashradius.value; // Alan!
	bolt->methodOfDeath = MOD_GRENADE;
	bolt->splashMethodOfDeath = MOD_GRENADE_SPLASH;
	bolt->clipmask = MASK_SHOT;
	
	bolt->s.pos.trType = TR_GRAVITY;
	bolt->s.pos.trTime = level.time - MISSILE_PRESTEP_TIME;         // move a bit on the very first frame
	VectorCopy( start, bolt->s.pos.trBase );
	/* Alan! change from 700 to g_grenade_speed.value */
	VectorScale( dir, g_grenade_speed.value, bolt->s.pos.trDelta );
	/* End Alan! change from 2k to g_plasma_speed.value */
	SnapVector( bolt->s.pos.trDelta );                      // save net bandwidth
	
	VectorCopy (start, bolt->r.currentOrigin);
	
	return bolt;
}

//=============================================================================


/*
=================
fire_bfg
=================
*/
gentity_t *fire_bfg (gentity_t *self, vec3_t start, vec3_t dir) {
	gentity_t       *bolt;
	
	VectorNormalize (dir);
	
	bolt = G_Spawn();
	bolt->classname = "bfg";
	bolt->nextthink = level.time + g_bfg_lifespan.value; // Alan! Use _lifespan.value instead of a constant!
	bolt->think = G_ExplodeMissile;
	bolt->s.eType = ET_MISSILE;
	bolt->r.svFlags = SVF_USE_CURRENT_ORIGIN;
	bolt->s.weapon = WP_BFG;
	bolt->r.ownerNum = self->s.number;
	bolt->parent = self;
	bolt->damage = g_bfg_damage.value; // Alan! using variables for funness...
	bolt->splashDamage = g_bfg_splashdamage.value; // Alan!
	bolt->splashRadius = g_bfg_splashradius.value; // Alan!
	bolt->methodOfDeath = MOD_BFG;
	bolt->splashMethodOfDeath = MOD_BFG_SPLASH;
	bolt->clipmask = MASK_SHOT;
	
	bolt->s.pos.trType = TR_LINEAR;
	bolt->s.pos.trTime = level.time - MISSILE_PRESTEP_TIME;         // move a bit on the very first frame
	VectorCopy( start, bolt->s.pos.trBase );
	/* Alan! change from 2k to g_bfg_speed.value */
	VectorScale( dir, g_bfg_speed.value, bolt->s.pos.trDelta );
	/* End Alan! change from 2k to g_bfg_speed.value */
	SnapVector( bolt->s.pos.trDelta );                      // save net bandwidth
	VectorCopy (start, bolt->r.currentOrigin);
	
	return bolt;
}

//=============================================================================


/*
=================
fire_rocket
=================
*/
gentity_t *fire_rocket (gentity_t *self, vec3_t start, vec3_t dir) {
	gentity_t       *bolt;
	
	VectorNormalize (dir);
	
	bolt = G_Spawn();
	bolt->classname = "rocket";
	bolt->nextthink = level.time + g_rocket_lifespan.value; // Alan! Use _lifespan.value instead of a constant!
	bolt->think = G_ExplodeMissile;
	bolt->s.eType = ET_MISSILE; 
	bolt->r.svFlags = SVF_USE_CURRENT_ORIGIN;
	bolt->s.weapon = WP_ROCKET_LAUNCHER;
	bolt->r.ownerNum = self->s.number;
	bolt->parent = self;
	bolt->damage = g_rocket_damage.value; // Alan! using variables for funness...
	bolt->splashDamage = g_rocket_splashdamage.value; // Alan!
	bolt->splashRadius = g_rocket_splashradius.value; // Alan!
	bolt->methodOfDeath = MOD_ROCKET;
	bolt->splashMethodOfDeath = MOD_ROCKET_SPLASH;
	bolt->clipmask = MASK_SHOT;
	
	bolt->s.pos.trType = TR_LINEAR;
	bolt->s.pos.trTime = level.time - MISSILE_PRESTEP_TIME;         // move a bit on the very first frame
	VectorCopy( start, bolt->s.pos.trBase );
	/* Begin Alan! Addition */
	VectorScale( dir, g_rocket_speed.value, bolt->s.pos.trDelta );
	/* End Alan! Add */
	
	SnapVector( bolt->s.pos.trDelta );                      // save net bandwidth
	VectorCopy (start, bolt->r.currentOrigin);
	
	return bolt;
}

/*
=================
fire_grapple
=================
*/
gentity_t *fire_grapple (gentity_t *self, vec3_t start, vec3_t dir) {
	gentity_t       *hook;
	
	VectorNormalize (dir);
	
	hook = G_Spawn();
	hook->classname = "hook";
	hook->nextthink = level.time + 10000;
	hook->think = Weapon_HookFree;
	hook->s.eType = ET_MISSILE;
	hook->r.svFlags = SVF_USE_CURRENT_ORIGIN;
	hook->s.weapon = WP_GRAPPLING_HOOK;
	hook->r.ownerNum = self->s.number;
	hook->methodOfDeath = MOD_GRAPPLE;
	hook->clipmask = MASK_SHOT;
	hook->parent = self;
	
	hook->s.pos.trType = TR_LINEAR;
	hook->s.pos.trTime = level.time - MISSILE_PRESTEP_TIME;         // move a bit on the very first frame
	hook->s.otherEntityNum = self->s.number; // use to match beam in client
	VectorCopy( start, hook->s.pos.trBase );
	VectorScale( dir, 800, hook->s.pos.trDelta );
	SnapVector( hook->s.pos.trDelta );                      // save net bandwidth
	VectorCopy (start, hook->r.currentOrigin);
	
	self->client->hook = hook;
	
	return hook;
}


/* Alan! Helper Functions */


void alanModifyPath( gentity_t *ent )
{
	float base;
	float param;
	int type;
	// Get the params needed for the current weapon.  I only check the first letter, so if a pulse rifle or 
	// something like that is put in, then this'll break.  But that's cool, because I don't care.
	switch( ent->classname[0] ) {
	case 'r': base = g_rocket_speed.value;  param = g_rocket_pathparam1.value;  type = g_rocket_pathfunc.value;  break;
	case 'p': base = g_plasma_speed.value;  param = g_plasma_pathparam1.value;  type = g_plasma_pathfunc.value;  break;
	case 'g': base = g_grenade_speed.value; param = g_grenade_pathparam1.value; type = g_grenade_pathfunc.value; break;
	case 'b': base = g_bfg_speed.value;     param = g_bfg_pathparam1.value;     type = g_bfg_pathfunc.value;	 break;
	}
	
	// -1 means to get the default value, so do that.
	if (param==-1) param = g_default_weapon_pathparam1.value;
	if (type==-1)  type = g_default_weapon_pathfunc.value;
	
	// Mmmmmmm... switch statement. 1 is the sine wave, that's all right now. */
	switch(type) {
	case 0:
		break;
	case 1:
		weaponSine( ent, base, param );
		break;
		// Add in any other weapon effects here.
		
	}
}

void weaponSine( gentity_t *ent, float base, float param ) {
	vec3_t vlast;
	vec3_t vnew;
	
	float s;
	
	// Calculate the last velocity change done. 
	s = sin( 2*M_PI*(level.previousTime/1000.0) )*(base*param);
	
	// Undo the last change made in the last click.  This is a little crappy because I didn't bother
	//      to check to see if this is the first tick.  Whatever, it's not noticable.  The actual code
	//      is just algebra to undo the addition of a perpendicular vector.  Blah.
	
#define V (ent->s.pos.trDelta)
	vlast[2] = ( V[2] - V[1]*s/base - V[0]*s*s/(base*base) )/(1.0-((s*s*s)/(base*base*base)));
	vlast[1] = V[1] + V[0]*s/base + vlast[2]*s*s/(base*base);
	vlast[0] = V[0] - vlast[2]*s/base;
	
	// Calculate the perpendicular vector.  Gives it at a crazy angle, but I like it that way,
	//      and this is easy. 
	vnew[1] = -vlast[0];
	vnew[2] =  vlast[1];
	vnew[0] =  vlast[2];
	
	// Normalize, probably could cheat and just divide by what the speed is supposed to be (base) 
	//      but, what fun would that be? 
	VectorNormalize( vnew );
	
	// S is now the change in velocity, so multiply by vnew and now you have a perpendicular
	//      to change the velocity! 
	s = sin( 2*M_PI*(level.time/1000.0) )*(base*param/100.0);
	VectorScale( vnew, s, vnew );
	
	// Add 'ems up and you're all done. 
	VectorAdd( vlast , vnew, ent->s.pos.trDelta );
	
	SnapVector( ent->s.pos.trDelta );                      // save net bandwidth (Or atleast that's what Mr. Carmack CLAIMS)
	
}

